##################################################################### imports

import matplotlib.pyplot
import networkx
import numpy
import random
import scipy
import scipy.ndimage
#import PIL
#import PIL.Image
import scipy.stats
import img2net_help
reload(img2net_help)

##################################################################### functions

def help_periodicdistance(lx,x1,x2,pbc):
    if(pbc==0):
        dx=numpy.abs(x1-x2)
    else:
        dx=numpy.array([numpy.abs(x1-x2),numpy.abs(x1-x2+lx),numpy.abs(x1-x2-lx)]).min()
    return dx

def help_periodiclist(lx,x1,pbc):
    if(pbc==0):
        dx=numpy.abs(numpy.subtract(range(lx),x1))
    else:
        dx=numpy.abs(numpy.minimum(numpy.minimum(numpy.abs(numpy.subtract(range(lx),x1)),numpy.abs(numpy.subtract(range(lx),lx+x1))),numpy.abs(numpy.subtract(range(lx),-lx+x1))))
    return dx

def help_edgekernel(lx,ly,vax,vay,x1,y1,x2,y2,pbcx,pbcy):
    dx1=help_periodiclist(lx,x1,pbcx)
    dx2=help_periodiclist(lx,x2,pbcx)
    dy1=help_periodiclist(ly,y1,pbcy)
    dy2=help_periodiclist(ly,y2,pbcy)
    ex1=numpy.ones((ly,1))*dx1**2/(2.0*vay)
    ey1=numpy.transpose(numpy.ones((lx,1))*dy1**2/(2.0*vax))
    ex2=numpy.ones((ly,1))*dx2**2/(2.0*vay)
    ey2=numpy.transpose(numpy.ones((lx,1))*dy2**2/(2.0*vax))
    ek=numpy.exp(-numpy.sqrt(ex1+ey1)-numpy.sqrt(ex2+ey2))
    return numpy.divide(ek,numpy.sum(ek))

def help_angle(vec):
    if(vec[0]<0):
        vec=-vec
    angle=180.0/numpy.pi*numpy.arccos(vec.dot([0,1])/numpy.sqrt(vec.dot(vec)))
    return angle

def grid_pos2D(N,nz,pos):
    pos2D={}
    pos3D={}
    for n in range(N/nz):
        pos2D[n]=(pos[n][0],pos[n][1])
    for n in range(N):
        pos3D[n]=(pos[n][0],pos[n][1])        
    return pos2D,pos3D

def grid_grid(crd,lx,ly,lz,dx,dy,dz,pbx,pby,pbz,vax,vay,compute_convs,dir_posi,dir_conv):
    
    pos={}
    edges=[]
    h=numpy.sqrt(3.0)/2.0
    dh=dy*h
    nx=int(lx/dx)
    ny=int(ly/dy)
    nh=int(ly/dh)
    nz=int(lz)
    DX=0.5*(lx-(nx-1)*dx)
    DY=0.5*(ly-(ny-1)*dy)
    DH=0.5*(ly-(nh-1)*dh)
    
    if(crd=='rectangular'):
        N=nx*ny*nz
        n=0
        for z in range(nz):
            for y in range(ny):
                for x in range(nx):
                    pos[n]=(x*dx+DX,y*dy+DY,z*dz)
                    n+=1
        dh=dy

    elif(crd=='triangular'):
        nx=nx-1+numpy.mod(nx,2)
        nh=nh-numpy.mod(nh,2)
        N=nx*nh*nz
        n=0
        for z in range(nz):
            for h in range(nh):
                for x in range(nx):
                    pos[n]=((x+0.5*numpy.mod(h,2))*dx+0.5*DX,(h+0.5)*dh+DH,z*dz)
                    n+=1
        ny=nh

    elif(crd=='hexagonal'):
        nx=3*int(nx/3)
        nh=nh-numpy.mod(nh,2)
        N=nx*nh*nz
        n=0
        for z in range(nz):
            for h in range(nh):
                for x in range(nx):
                    if(numpy.mod(h,2)==0 and numpy.mod(x+2,3)!=0):
                        pos[n]=((x+0.5+0.5*numpy.mod(h,2))*dx+0.5*DX,(h+0.5)*dh+DH,z*dz)
                        n+=1
                    if(numpy.mod(h,2)==1 and numpy.mod(x+1,3)!=0):
                        pos[n]=((x+0.5+0.5*numpy.mod(h,2))*dx+0.5*DX,(h+0.5)*dh+DH,z*dz)
                        n+=1
        N=n
        ny=nh

    for n in range(N):
        for m in range(n):
            Dx=help_periodicdistance(nx*dx,pos[n][0],pos[m][0],pbx)
            Dy=help_periodicdistance(ny*dh,pos[n][1],pos[m][1],pby)
            Dz=help_periodicdistance(nz*dz,pos[n][2],pos[m][2],pbz)
            if((Dx/dx)**2+(Dy/dy)**2+(Dz/dz)**2<1.1):
                edges.append((n,m))

    E=len(edges)    
    pos2D,pos3D=grid_pos2D(N,nz,pos)

    convs=[]
    if(compute_convs):
        numpy.save(dir_posi,pos)            
        for e in range(E):
            n=edges[e][0]
            m=edges[e][1]
            x1=pos3D[n][0]
            y1=pos3D[n][1]
            x2=pos3D[m][0]
            y2=pos3D[m][1]
            conv=help_edgekernel(lx,ly,vax,vay,x1,y1,x2,y2,pbx,pby)
            numpy.save(dir_conv+'_L='+str(e).zfill(4),conv)
            if(E<1000):
                convs.append(conv)

    return N,nx,ny,nz,pos,E,edges,convs
    


def grid_graph(im,E,edges,dir_pos,dir_convs,dz):
    capas=[]
    pos=numpy.load(dir_pos+'.npy').flatten()[0]    
    for e in range(E):
        conv=numpy.load(dir_convs+'_L='+str(e).zfill(4)+'.npy')    
        n=edges[e][0]
        m=edges[e][1]
        z0=int(pos[n][2]/dz)
        z1=int(pos[m][2]/dz)
        if(z0!=z1):
            capas.append(numpy.sum(numpy.multiply(0.5*(im[z0]+im[z1]),conv)))
        else:
            capas.append(numpy.sum(numpy.multiply(im[z0],conv)))
    capas=numpy.divide(capas,numpy.sum(capas))

    graph=networkx.Graph()
    for e in range(E):
        n=edges[e][0]
        m=edges[e][1]
        graph.add_edge(n,m,capa=capas[e],lgth=1.0/capas[e])
    return graph

def grid_all(inp):    
    t,e,i,directory,files,L,edges,dir_pos,dir_convs,dz=inp    
    ims=[]
    for z in range(len(files)):
        im=1.0*scipy.ndimage.imread(directory+files[z])
#        im=1.0*numpy.array(PIL.Image.open(directory+files[z]))
        if(len(numpy.shape(im))>2):
            im=im[:,:,0]
        ims.append(im)          
    graph=img2net_help.grid_graph(ims,L,edges,dir_pos,dir_convs,dz)
    
    return t,e,i,graph

def graph_null(graphn,nx,ny,numo):
        graphg=graphn.copy()

        if(numo=='edges'):
            E=graphg.number_of_edges()
            idx=range(E)
            random.shuffle(idx)
            for e in range(E):
                graphg.edges(data=1)[e][2]['capa']=graphn.edges(data=1)[idx[e]][2]['capa']
                graphg.edges(data=1)[e][2]['lgth']=graphn.edges(data=1)[idx[e]][2]['lgth']

        elif(numo=='lines'):
            E=graphg.number_of_edges()
            idx=range(nx)
            idy=range(ny)
            random.shuffle(idx)
            random.shuffle(idy)
            links=graphg.edges(data=1)
            for e in range(E):
                [n0,n1]=links[e][0:2]
                [p0x,p0y]=[numpy.mod(n0,nx),n0/nx]
                [p1x,p1y]=[numpy.mod(n1,nx),n1/nx]
                if(p0x==p1x):
                    graphg.edges(data=1)[e][2]['capa']=graphn[idx[p0x]+p0y*nx][idx[p1x]+p1y*nx]['capa']
                    graphg.edges(data=1)[e][2]['lgth']=graphn[idx[p0x]+p0y*nx][idx[p1x]+p1y*nx]['lgth']
                if(p0y==p1y):
                    graphg.edges(data=1)[e][2]['capa']=graphn[p0x+idy[p0y]*nx][p1x+idy[p1y]*nx]['capa']
                    graphg.edges(data=1)[e][2]['lgth']=graphn[p0x+idy[p0y]*nx][p1x+idy[p1y]*nx]['lgth']

        elif(numo=='blocks'):
            Bx,By=3,4
            graphg=graphn.copy()
            E=graphg.number_of_edges()
            N=graphg.number_of_nodes()
            bx=nx/Bx
            by=ny/By
            B=Bx*By
            idb=range(B)
            random.shuffle(idb)
            links=graphg.edges(data=1)
            left=set(range(E))
            for e in range(E):
                [n0,n1]=links[e][0:2]
                [p0x,p0y]=[numpy.mod(n0,nx),n0/nx]
                [p1x,p1y]=[numpy.mod(n1,nx),n1/nx]
                [c0x,c0y]=[p0x/bx,p0y/by]
                [c1x,c1y]=[p1x/bx,p1y/by]
                [c0,c1]=[c0x+c0y*Bx,c1x+c1y*Bx]
                if(c0==c1 and c0<B):
                    left=left.difference([e])
                    [cnewx,cnewy]=[numpy.mod(idb[c0],Bx),idb[c0]/Bx]
                    [d0x,d0y]=[p0x+(cnewx-c0x)*bx,p0y+(cnewy-c0y)*by]
                    [d1x,d1y]=[p1x+(cnewx-c1x)*bx,p1y+(cnewy-c1y)*by]
                    [m0,m1]=[d0x+d0y*nx,d1x+d1y*nx]
                    graphg.edges(data=1)[e][2]['capa']=graphn[m0][m1]['capa']
                    graphg.edges(data=1)[e][2]['lgth']=graphn[m0][m1]['lgth']
            left=list(left)
            ew=[graphg.edges(data=1)[i][2]['capa'] for i in left]
            el=[graphg.edges(data=1)[i][2]['lgth'] for i in left]
            random.shuffle(left)
            for e,ee in enumerate(left):
                graphg.edges(data=1)[ee][2]['capa']=ew[e]
                graphg.edges(data=1)[ee][2]['lgth']=el[e]

        return graphg

def graph_all(inp):
    [t,e,i,r,G,pos]=inp
    data=[]
    label=[]
    N=G.number_of_nodes()

    ################################################# degree
    deg=G.degree(weight='capa').values()

    label.append('mean[degree]')
    data.append(numpy.mean(deg))
    label.append('sd[degree]')
    data.append(numpy.std(deg))
    label.append('skewness[degree]')
    data.append(scipy.stats.skew(deg))

    ################################################# structure

    label.append('clustering coefficient')
    data.append(networkx.average_clustering(G,weight='capa'))
    label.append('assortativity')
    data.append(networkx.degree_pearson_correlation_coefficient(G,weight='capa'))

    ################################################# distances
    dists=networkx.all_pairs_dijkstra_path_length(G,weight='lgth')
    dist=[[v for v in u.itervalues()] for u in dists.itervalues()]
    ecce=numpy.array(dist).max(0)

    label.append('mean[distance]')
    data.append(numpy.mean(dist))
    label.append('sd[distance]')
    data.append(numpy.std(dist))
    label.append('skewness[distance]')
    data.append(scipy.stats.skew(numpy.reshape(dist,-1)))
    label.append('radius')
    data.append(ecce.min())
    label.append('diameter')
    data.append(ecce.max())

    ################################################# eigenvalues
    spec=numpy.sort(networkx.laplacian_spectrum(G,weight='capa'))

    label.append('effective resistance')
    data.append(1.0/numpy.sum(numpy.divide(1.0*N,spec[1:N-1])))
    label.append('algebraic connectivity')
    data.append(spec[1])

    ################################################# betweenness
    flow=networkx.edge_current_flow_betweenness_centrality(G,weight='capa',normalized=1).values()

    label.append('mean[betweenness]')
    data.append(numpy.mean(flow))
    label.append('sd[betweenness]')
    data.append(numpy.std(flow))
    label.append('skewness[betweenness]')
    data.append(scipy.stats.skew(flow))

    ################################################# angles
    angle_angle=[]
    angle_weight=[]
    for u,v,d in G.edges_iter(data=1):
        angle_angle.append(img2net_help.help_angle(numpy.subtract(pos[u][0:2],pos[v][0:2])))
        angle_weight.append(d['capa'])
    angle_angle=numpy.mod(angle_angle,180)

    label.append('angle 000')
    data.append(numpy.sum(numpy.where(angle_angle==0,angle_weight,0))/numpy.sum(numpy.where(angle_angle==0,1,0)))  ### ACHTUNG normierung?
    label.append('angle 045')
    data.append(numpy.sum(numpy.where(angle_angle==45,angle_weight,0))/numpy.sum(numpy.where(angle_angle==45,1,0)))
    label.append('angle 060')
    data.append(numpy.sum(numpy.where(angle_angle==60,angle_weight,0))/numpy.sum(numpy.where(angle_angle==60,1,0)))
    label.append('angle 090')
    data.append(numpy.sum(numpy.where(angle_angle==90,angle_weight,0))/numpy.sum(numpy.where(angle_angle==90,1,0)))
    label.append('angle 120')
    data.append(numpy.sum(numpy.where(angle_angle==120,angle_weight,0))/numpy.sum(numpy.where(angle_angle==120,1,0)))
    label.append('angle 135')
    data.append(numpy.sum(numpy.where(angle_angle==135,angle_weight,0))/numpy.sum(numpy.where(angle_angle==135,1,0)))
    label.append('angle ratio 00-90')
    data.append(data[-6]/data[-3])

    return t,e,i,r,data,label


def boxplot(data,color,labels,ttest):
    L=len(labels)
    bp=matplotlib.pyplot.boxplot(data,sym='',notch=0,widths=0.5)
    [matplotlib.pyplot.setp(bp[k],color=color,ls='-',alpha=1.0,lw=2.0) for k in bp.keys()]
    lims=numpy.array([bp['whiskers'][l].get_data()[1] for l in range(2*L)])  
    matplotlib.pyplot.xticks(range(1,1+L),labels)
    limi,lima=matplotlib.pyplot.ylim()
    limd=lima-limi
    matplotlib.pyplot.ylim([limi,lima+0.1*limd])
    if(ttest==1):        
        for i in range(len(labels)):
            pval=scipy.stats.ttest_1samp(data[i],1)[1]
            matplotlib.pyplot.text(i+1,lima+0.03*limd,"p=%.2f"%pval,ha='center')    
    return None
    
